/********************************************************************************************/
/*   LOGIC ANALYZER                                                                         */
/********************************************************************************************/
/* Author: Raul Navarro                                                                     */
/* Web:    http://www.raulnd.com                                                            */
/* E-Mail: raulnd@gmail.com                                                                 */
/********************************************************************************************/
/*                                                                                          */
/* This program is free software: you can redistribute it and/or modify                     */
/* it under the terms of the GNU General Public License as published by                     */
/* the Free Software Foundation, either version 3 of the License, or                        */
/* (at your option) any later version.                                                      */
/*                                                                                          */
/* This program is distributed in the hope that it will be useful,                          */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of                           */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                            */
/* GNU General Public License for more details.                                             */
/*                                                                                          */
/* You should have received a copy of the GNU General Public License                        */
/* along with this program.  If not, see  < http://www.gnu.org/licenses/ >.                 */
/*                                                                                          */
/********************************************************************************************/

// Arduino UNO and Arduino Nano 328 have 2K SRAM.
// Use 1800 bytes for data buffer and spare 248.
// Analog data: 1800 / 2 =  900 scans

#define DATA 1800

/*** From wiring_private.h ***/

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif

#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif 



/*****************************/

int i;
int a;
char c;
unsigned long t;

byte _data[DATA];

void setup()
{
  
  /*** From wiring.c ***/
  
  // set a2d prescale factor to 128
  // 16 MHz / 128 = 125 KHz, inside the desired 50-200 KHz range.
  // XXX: this will not work properly for other clock speeds, and
  // this code should use F_CPU to determine the prescale factor.
  //sbi(ADCSRA, ADPS2);
  //sbi(ADCSRA, ADPS1);
  //sbi(ADCSRA, ADPS0);
  
  // Datasheet: http://www.atmel.com/Images/doc8161.pdf
  
  // Page: 263.
  // 23.9.2 ADCSRA – ADC Control and Status Register A
  //     ADPS2 ADPS1 ADPS0
  // BIT     2     1     0
  
  // Page: 264.
  // Table: 23-5.
  // ADPS2 ADPS1 ADPS0 DIVISOR
  //     1     0     0      16

  // set a2d prescale factor to 16
  sbi(ADCSRA, ADPS2);
  cbi(ADCSRA, ADPS1);
  cbi(ADCSRA, ADPS0);
  
  /*********************/
  
  // DEFAULT Analog reference + A0 input pin
  ADMUX = (DEFAULT << 6); // | (pin & 0x07);
  
  for(int i=8; i<=13; i++)
  {
    pinMode(i, INPUT);
    digitalWrite(i, LOW);
  }
  
  Serial.begin(115200);
  
}

void loop()
{
  
  if(Serial.available() > 0)
  {
    c = Serial.read();
    switch(c)
    {
    
      case 'D':
      
        BeginRead();
        ReadDigital();
        EndRead();
        
        break;
      
      case 'A':

        BeginRead();
        ReadAnalog();
        EndRead();
      
        break;
     
    }
    
  }

}

void BeginRead()
{
  Serial.end();
  noInterrupts();  
}

void EndRead()
{
  interrupts();
  Serial.begin(115200);
  Serial.write(_data, DATA);
}

/*******************************************/
/* 8-bit AVR® Instruction Set              */
/* http://www.atmel.com/Images/doc0856.pdf */
/*                                         */
/* IN Rd, A       1 Cycle           PAG 13 */
/* ST Z+, Rr      2 Cycles          PAG 13 */
/*                                         */
/* 3 Cycles * 1800 Times = 5400 Cycles     */
/* Xtal 16 MHz = 62.5 ns per cycle         */
/* 5400 Cycles * 62.5 ns = 337.5 us        */
/* 1800 times / 337.5 us = 5.33 MHz        */
/*******************************************/

// Do not use loops because they consume several clock cycles.
// 1800 times IN and ST.
void ReadDigital()
{
  __asm__ __volatile(
#include "asm.h";
    :
    : [port] "i" (_SFR_IO_ADDR(PINB)), "z" (_data)
    :
  );
}

/*******************************************/
/* PRESCALER 16                            */
/* Read = 215 cycles                       */
/* 215 Cycles * 900 Times = 193500 Cycles  */
/* Xtal 16 MHz = 62.5 ns per cycle         */
/* 193500 Cycles * 62.5 ns = 12.09 ms      */
/* 900 times / 12.09.5 ms = 13.44 KHz      */
/*******************************************/

void ReadAnalog()
{

  for(i=0; i<DATA; i+=2) // 2 cycles
  {

    /*** From wiring_analog.c ***/

    // start the conversion
    sbi(ADCSRA, ADSC); // 2 Cycles

    // ADSC is cleared when the conversion finishes
    while (bit_is_set(ADCSRA, ADSC)); // 13 * 16 Cycles

    // we have to read ADCL first; doing so locks both ADCL
    // and ADCH until ADCH is read.  reading ADCL second would
    // cause the results of each conversion to be discarded,
    // as ADCL and ADCH would be locked when it completed.
    _data[i+1] = ADCL; // 2 cycles
    _data[i]   = ADCH; // 1 cycle

    /****************************/
    
  }
  
}

